/*
	NumInputBox	 - handy form fields with labels, statusBar description,
					numeric parsing, and so on

*/

package mylib;

import java.util.*;
import java.awt.*;
import java.applet.*;
import mylib.MyFuns;
import mylib.DescribedTextArea;
import mylib.DescribedTextField;
import java.net.*;
import java.io.*;
import java.io.DataOutputStream;
import netscape.security.PrivilegeManager;

class AddPageButton extends Button
{
  NumInputBox parent;

  AddPageButton (NumInputBox parent)
    {
      super (" New Page ");
      this.parent = parent;
    }

  public boolean action(Event  evt, Object  what)
    {
      parent.addPage ();
      return true;
    }
}

class PasteFromVCButton extends Button
{
  NumInputBox parent;
	
  PasteFromVCButton (NumInputBox parent)
    {
      super (" <- VC ");
      this.parent = parent;
    }

  public boolean action(Event  evt, Object  what)
    {
      parent.pasteFromVC ();
      return true;
    }
}

class CopyToVCButton extends Button
{
  NumInputBox parent;
	
  CopyToVCButton (NumInputBox parent)
    {
      super (" -> VC ");
      this.parent = parent;
    }

  public boolean action(Event  evt, Object  what)
    {
      parent.copyToVC ();
      return true;
    }
}

class LoadButton extends Button
{
  NumInputBox parent;
	
  LoadButton (NumInputBox parent)
    {
      super (" Load ");
      this.parent = parent;
    }

  public boolean action(Event  evt, Object  what)
    {
      parent.loadFromFile ();
      return true;
    }
}

class SaveButton extends Button
{
  NumInputBox parent;
	
  SaveButton (NumInputBox parent)
    {
      super (" Save ");
      this.parent = parent;
    }

  public boolean action(Event  evt, Object  what)
    {
      parent.saveToFile ();
      return true;
    }
}

class PageChoice extends Choice
{
  NumInputBox parent;

  public PageChoice (NumInputBox parent)
    {
      super ();
      this.parent = parent;
    }

  public boolean handleEvent (Event  evt)
    {
      if (evt.id == Event.ACTION_EVENT)
	{
	  parent.gotoPage (getSelectedIndex ());
	  return true;
	}
      return false;
    }
}



public class NumInputBox
{
  // flag: Virtual Clipboard/Page buttons are put to the right of 
  //		the input box if buttonsOnSide is true, otherwise they are put below

  static boolean buttonsOnSide = true;

  String title;
  String statusBarInfo;
  int dim;					// 0, 1, 2 - dimensions of the input box
  public TextComponent tf;	// either a TextField or a TextArea
  Container parent;				// gives access to the statusbar
  Label lab;

  static final int MAXPAGESIZE = 10000;

	// currently, the textcomponents of Java seem to impose a ~30000 character
	// limit on the amount of text in a component, so we add
	// a bank-switching button
	// Actually, Netscape on the Mac imposes around 12000, but 30000 on the PC

  String bigBuff = "";				// all pages are stored in here
  int pageStarts[] = new int[20];		// a vector of integers to represent indexes into bigBuff of page-strings (this vector will be expanded if necessary)
  int numPages = 1;
  int currPage = 0;
  PageChoice			chPage;
  Label				labPage;
  AddPageButton		btnAddPage;			// button to add new page to window
  PasteFromVCButton	btnPasteFromVC;
  CopyToVCButton		btnCopyToVC;
  LoadButton   btnLoad;
  SaveButton   btnSave;
  int pageLocX, pageLocY;					// location of page chooser, for later replacement

  private String pageLabel ()
    {
      return " of " +  numPages + " Page(s)        ";
    }
	
  public static void setButtonsOnSide (boolean b)
    {
      buttonsOnSide = b;
    }

  public NumInputBox
  (
		
   Container a, String t, String s, int d,
   int grid_x, int grid_y, int grid_width, int grid_height,
   int fill, int anchor, double weight_x, double weight_y,
   int top, int left, int bottom, int right
   )
    {
      this (a, t, s, d, grid_x, grid_y, grid_width, grid_height, fill, anchor,  weight_x, weight_y, top, left, bottom, right, true);
    }

  public NumInputBox
  (
		
   Container a, String t, String s, int d,
   int grid_x, int grid_y, int grid_width, int grid_height,
   int fill, int anchor, double weight_x, double weight_y,
   int top, int left, int bottom, int right, boolean isInput
   )
    {
      Font f = new Font("Courier", 0, 14);
      parent = a;
      title = t;
      statusBarInfo = s;
      dim = d;
      lab = new Label (" " + t + " ", Label.LEFT);
      lab.setFont (new Font("Times", Font.BOLD, 14));
      switch (dim)
	{
	case 0:
	  tf = new DescribedTextField (a, s, 8);
	  tf.setFont (f);
	  break;

	case 1:
	  tf = new DescribedTextArea (a, s, 8, 8);
	  tf.setFont (f);
	  break;

	default:
	  tf = new DescribedTextArea (a, s, 8, 40);
	  tf.setFont (f);
	  labPage = new Label (pageLabel (), Label.LEFT);
	  chPage = new PageChoice (this);
	  chPage.addItem ("1");
	  labPage.setText (pageLabel());
	  pageStarts[0] = 0;
	  btnCopyToVC = new CopyToVCButton (this);
	  btnSave = new SaveButton (this);
	  if (isInput)
	    {
	      btnAddPage = new AddPageButton (this);
	      btnPasteFromVC = new PasteFromVCButton (this);
	      btnLoad = new LoadButton (this);
	    }
	  break;
	}
      switch (dim)
	{
	case 0:
	case 1:
	  MyFuns.constrain (a, lab, grid_x, grid_y, grid_width, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.SOUTHWEST);
	  MyFuns.constrain (a, tf, grid_x, grid_y+1, grid_width, grid_height, fill, anchor, weight_x, weight_y, top, left, bottom, right );
	  break;
	default:
	  MyFuns.constrain (a, lab, grid_x, grid_y, 2, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.WEST);
	  int dx = 2;
	  int dy = 1;
	  if (buttonsOnSide)
	    {
	      if (isInput)
		{
		  MyFuns.constrain (a, btnAddPage, grid_x + 5, grid_y+dy++, 1, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.WEST );
		  MyFuns.constrain (a, btnPasteFromVC, grid_x + 5, grid_y+dy++, 1, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.WEST);
		  MyFuns.constrain (a, btnLoad, grid_x+5, grid_y+dy--, 1, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.WEST);
		}
	      MyFuns.constrain (a, btnCopyToVC, grid_x + 6, grid_y+dy++, 1, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.WEST );
	      MyFuns.constrain (a, btnSave, grid_x + 6, grid_y+dy++, 1, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.WEST );
	    }
	  else
	    {
	      if (isInput)
		{
		  MyFuns.constrain (a, btnAddPage, grid_x, grid_y + 5, 1, 1, GridBagConstraints.NONE, GridBagConstraints.NORTH );
		  MyFuns.constrain (a, btnPasteFromVC, grid_x + 1, grid_y + 5, 1, 1, GridBagConstraints.NONE, GridBagConstraints.NORTH );
		  MyFuns.constrain (a, btnLoad, grid_x + 3, grid_y + 5, 1, 1, GridBagConstraints.NONE, GridBagConstraints.NORTH );
		}
	      MyFuns.constrain (a, btnCopyToVC, grid_x + 2, grid_y + 5, 1, 1, GridBagConstraints.NONE, GridBagConstraints.NORTH);
	      MyFuns.constrain (a, btnSave, grid_x + 4, grid_y + 5, 1, 1, GridBagConstraints.NONE, GridBagConstraints.NORTH);
	    }
	  MyFuns.constrain (a, chPage, pageLocX = (grid_x + dx++), pageLocY = grid_y, 1, 1, GridBagConstraints.NONE, GridBagConstraints.SOUTHEAST );
	  MyFuns.constrain (a, labPage, grid_x + dx++, grid_y, 1, 1, GridBagConstraints.NONE, GridBagConstraints.SOUTHWEST );
	  MyFuns.constrain (a, tf, grid_x, grid_y+1, grid_width, grid_height+3, fill, anchor, weight_x, weight_y, top, left, bottom, right );
	  break;
	}
    }

  public NumInputBox (
		      
		      Container a, String t, String s, int d,
		      int grid_x, int grid_y, int grid_width, int grid_height,
		      int fill, int anchor
		      )  
    {
      this (a, t, s, d, grid_x, grid_y, grid_width, grid_height, fill, anchor, 0.0, 0.0, 0, 0, 0, 0, true);
    }

public NumInputBox (

 Container a, String t, String s, int d,
 int grid_x, int grid_y, int grid_width, int grid_height,
 int fill, int anchor, boolean isInput
 )

{
  this (a, t, s, d, grid_x, grid_y, grid_width, grid_height, fill, anchor, 0.0, 0.0, 0, 0, 0, 0, isInput);
}

public NumInputBox
(
		
 Container a, String t, String s, int d,
 int grid_x, int grid_y, int grid_width, int grid_height,
 int top, int left, int bottom, int right
 )

{
  this (a, t, s, d, grid_x, grid_y, grid_width, grid_height, GridBagConstraints.NONE, GridBagConstraints.CENTER,  0.0, 0.0, top, left, bottom, right, true);
}

public NumInputBox
(
		
 Container a, String t, String s, int d,
 int grid_x, int grid_y, int grid_width, int grid_height,
 int top, int left, int bottom, int right, boolean isInput
 )

{
  this (a, t, s, d, grid_x, grid_y, grid_width, grid_height, GridBagConstraints.NONE, GridBagConstraints.CENTER,  0.0, 0.0, top, left, bottom, right, isInput);
}

public NumInputBox
(
		
 Container a, String t, String s, int d,
 int grid_x, int grid_y, int grid_width, int grid_height
 )

{
  this (a, t, s, d, grid_x, grid_y, grid_width, grid_height, GridBagConstraints.NONE, GridBagConstraints.CENTER,  0.0, 0.0, 0, 0, 0, 0, true);
}

public NumInputBox
(
		
 Container a, String t, String s, int d,
 int grid_x, int grid_y, int grid_width, int grid_height,
 boolean isInput
 )

{
  this (a, t, s, d, grid_x, grid_y, grid_width, grid_height, GridBagConstraints.NONE, GridBagConstraints.CENTER,  0.0, 0.0, 0, 0, 0, 0, isInput);
}

public final int intValue () throws Exception
{
  try
    {
      return Double.valueOf ( getText () ) . intValue ();
    }
  catch (Exception e)
    {
      throw new Exception ("Can't get an integer from " + title + ".\n" + e.getMessage () );
    }
}

public final int posIntValue () throws Exception
{
  try
    {
      int i = Double.valueOf ( getText () ) . intValue ();
      if (i > 0)
	return i;
      else
	throw new Exception ("Must be greater than zero.");
    }
  catch (Exception e)
    {
      throw new Exception ("Can't get an integer from " + title + ".\n" + e.getMessage () );
    }
}

public String getText ()
{
  if (dim > 1)
    {
      // copy the values from the current page into bigBuff
      gotoPage (currPage);
      return bigBuff;
    }
  else
    {
      return tf.getText ();
    }
}

public final int[] getNumValues (int a[]) throws Exception
{
  try 
    {
      if (a == null)
	a = MyFuns.StringToInts ( getText () );
      else
	MyFuns.StringToNumbers ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting an integer from " + title + ".\n" + e.getMessage () );
    }
  return a;
}

public final void getNumValues (int a[][], int sizes[]) throws Exception
{
  try 
    {
      if (sizes != null)
	MyFuns.AllocVectors (a, sizes);
      MyFuns.StringToNumbers ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting integers from " + title + ".\n" + e.getMessage () );
    }
}

public final void getNumValues (int a[][][], int sizes[]) throws Exception
{
  try 
    {
      if (sizes != null)
	MyFuns.AllocVectors (a, sizes);
      MyFuns.StringToNumbers ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting integers from " + title + ".\n" + e.getMessage () );
    }
}


public final double doubleValue () throws Exception
{
  try
    {
      return Double.valueOf ( getText () ) . doubleValue ();
    }
  catch (Exception e)
    {
      throw new Exception ("Can't get a real number from " + title + ".\n" + e.getMessage () );
    }
}

public final double [] getNumValues (double a[]) throws Exception
{
  try 
    {
      if (a == null)
	a = MyFuns.StringToDoubles ( getText () );
      else
	MyFuns.StringToNumbers ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting a real number from " + title + ".\n" + e.getMessage () );
    }
  return a;
}

public final void getNumValues (double a[][], int sizes[]) throws Exception
{
  try 
    {
      if (sizes != null)
	MyFuns.AllocVectors (a, sizes);
      MyFuns.StringToNumbers ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting real numbers from " + title + ".\n" + e.getMessage () );
    }
}

public final void getNumValues (double a[][][], int sizes[]) throws Exception
{
  try 
    {
      if (sizes != null)
	MyFuns.AllocVectors (a, sizes);
      MyFuns.StringToNumbers ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting real numbers from " + title + ".\n" + e.getMessage () );
    }
}

public final int[] getLabelValues (int a[]) throws Exception
{
  try 
    {
      if (a == null)
	a = MyFuns.StringToHashCodes ( getText () );
      else
	MyFuns.StringToHashCodes ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting labels from " + title + ".\n" + e.getMessage () );
    }
  return a;
}

public final void getLabelValues (int a[][], int sizes[]) throws Exception
{
  try 
    {
      if (sizes != null)
	MyFuns.AllocVectors (a, sizes);
      MyFuns.StringToHashCodes ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting labels from " + title + ".\n" + e.getMessage () );
    }
}

public final void getLabelValues (int a[][][], int sizes[]) throws Exception
{
  try 
    {
      if (sizes != null)
	MyFuns.AllocVectors (a, sizes);
      MyFuns.StringToHashCodes ( a, getText (), true );
    }
  catch (Exception e)
    {
      throw new Exception ("Problem getting labels from " + title + ".\n" + e.getMessage () );
    }
}

public final int countValues ()
{
  return MyFuns.CountFieldsInString (getText ());
}

public final void setText (String s)
{
  if (! System.getProperty ("line.separator") .equals ("\r") )
    {
      s.replace ('\r', ' ');
    }

  if (dim < 2)
    {
      tf.setText (s);
    }
  else
    {

      ((DescribedTextArea)tf).setAltered (false);

      // only trick is to set page starting points
      // we try not to break a line across pages

      bigBuff = s;
      int l = s.length ();
      int prevNumPages = numPages;
      numPages = 0;

      int i = 0;
      while ( i <= l)
	{
	
	  addPageSlot ();	// increase number of pages
	  pageStarts[numPages-1] = i;
			
	  // try to break this page at the end of a line
	  // or else at whitespace, or else at one page's size
	  // (the last option is a problem since different pages
	  // are elsewhere treated as separated by a space, 
	  // but it is only taken when the string is huge

	  if (l - i < MAXPAGESIZE)
	    break;

	  int j = 1 + Math.max (s.lastIndexOf ('\n', i + MAXPAGESIZE - 1), s.lastIndexOf ('\r', i + MAXPAGESIZE - 1) );
	  if (j < i)		// if no linefeed found on this page
	    {
	      j = 1 + s.lastIndexOf (' ', i + MAXPAGESIZE);
	      if (j < i)
		j = 1 + s.lastIndexOf ('\t', i + MAXPAGESIZE);
	    }
	  i = (j > i) ? j : i + MAXPAGESIZE;
	}
			
      if (prevNumPages > numPages)
	{
	  parent.remove (chPage);

	  chPage = new PageChoice (this);

	  for (int c = 1; c <= numPages; ++c)
	    chPage.addItem (String.valueOf (c));

	  chPage.hide ();
	  MyFuns.constrain (parent, chPage, pageLocX, pageLocY, 1, 1, GridBagConstraints.NONE, GridBagConstraints.SOUTHWEST );
	  chPage.show ();
	  parent.layout ();
	}
      else
	{
	  while (prevNumPages < numPages)
	    chPage.addItem (String.valueOf (1 + prevNumPages++));
	}


      showPage (0);
    }
}

public final void showPage (int p)
{
  // display the contents of page p (p >= 0)
  // without worrying about what's showing in the window now
  if (p < numPages )
    {
      if (p < numPages - 1)
	tf.setText (bigBuff.substring (pageStarts [p], pageStarts [p+1]));
      else
	if (pageStarts[p] < bigBuff.length () )
	  tf.setText (bigBuff.substring (pageStarts [p]));
	else
	  tf.setText ("");
      labPage.setText (pageLabel ());
      currPage = p;
    }
}

public final void gotoPage (int p)
{
	
  // show a page, but if necessary, record the current window contents in its
  // proper position, adding an end-of-line to the window if there isn't
  // one (this guarantees whitespace between the last field in one page
  // and the first field on the next)

  int cp = currPage;
  String s = tf.getText ();
  int l = s.length ();
  if (l == 0 || (s.charAt (l-1) != '\n' && s.charAt (l-1) != '\r'))
    {
      s = s + '\n';
      ++l;
    }
  String newBuff = (cp == 0) ? "" : bigBuff.substring (0, pageStarts [cp]);
  newBuff = newBuff.concat (s);
  newBuff = (cp == numPages - 1 ) ? newBuff : newBuff.concat (bigBuff.substring (pageStarts [cp+1]));
  bigBuff = newBuff;
  if (cp < numPages - 1)
    {
      // adjust starts of subsequent pages, if necessary

      int deltaPageSize = l - (pageStarts [cp + 1] - pageStarts [cp]);
      for (int j = cp + 1; j < numPages; ++j)
	pageStarts [j] += deltaPageSize;
    }
  if (cp != p)
    {
      showPage (p);
    }
}

private final void addPageSlot ()
{
  if (++ numPages >= pageStarts.length )
    {
      int tmp[] = pageStarts;
      pageStarts = new int[pageStarts.length + 10];
      System.arraycopy (tmp, 0, pageStarts, 0, tmp.length);
    }
}

public final void addPage ()
{
  addPageSlot ();
  pageStarts [numPages - 1] = bigBuff.length ();
  //		bigBuff = bigBuff.concat (" ");	// put a space on the new page
  chPage.addItem (String.valueOf (numPages));
  chPage.select (numPages - 1);
  gotoPage (numPages - 1);
}

public final void pasteFromVC ()
{
  try
    {
      if (parent instanceof Applet)
	((Applet)parent).showStatus ("Pasting from Virtual Clipboard...");
      URLConnection conVC = (new URL ("http://gause.biology.ualberta.ca/jmbbin/viewclip")) .openConnection ();
      conVC.setUseCaches(false);
      conVC.setRequestProperty ("method", "POST");
      conVC.connect ();
      int size = conVC.getContentLength ();
      byte buf[] = new byte[size];
      InputStream is =  conVC.getInputStream ();
      int byteCount = 0;
      while (byteCount < size)
	{
	  int j = is.read (buf, byteCount, size - byteCount);
	  if (j <= 0)			// just in case server lied about content-length
	    break;
	  byteCount += j;
	}

      is.close ();
      if (! System.getProperty ("line.separator").equals ( "\r"))
	{
	  int i;
	  for (i=0; i < byteCount-1; ++i)
	    if (buf[i] == 13)
	      if (buf[i+1] == 10)
		buf[i] = 32;
	      else
		buf[i] = 10;

	  if (buf[i] == 13)
	    buf[i] = 10;
	}
      setText (new String (buf, 0, 0, byteCount));
      if (parent instanceof Applet)
	((Applet)parent).showStatus ("Pasted " + byteCount + " bytes from Virtual Clipboard.");
    }
  catch (Exception e)
    {
      if (parent instanceof Applet)
	((Applet)parent).showStatus ("Error pasting from Virtual Clipboard: " + e.getMessage ());
    }
}


public final void copyToVC ()
{
  try
    {
      if (parent instanceof Applet)
	((Applet)parent).showStatus ("Copying to Virtual Clipboard...");

      String s = getText ();
      int size = s.length ();
      Socket cbSock = new Socket ("gause.biology.ualberta.ca", 80, true);
      DataOutputStream dos =  new DataOutputStream (cbSock.getOutputStream ());
      if (dos == null)
	throw new Exception ("Couldn't open output stream.");

      dos.writeBytes ("POST /jmbbin/copytoclip HTTP/1.0\n");
      dos.writeBytes ("Content-Encoding: text/plain\n");
      dos.writeBytes ("Content-Length: " + (size + 151) + "\n\n");
      dos.writeBytes ("-----------------------------25643213361610\n");
      dos.writeBytes ("Content-Disposition: form-data; filename=\"_from_browser\"\n\n");
      dos.writeBytes (s);
      dos.writeBytes ("\n-----------------------------25643213361610--\n");
      dos.flush ();
      dos.close ();
      cbSock.close ();

      if (parent instanceof Applet)
	((Applet)parent).showStatus ("Copied " + size + " bytes to Virtual Clipboard.");
    }
  catch (Exception e)
    {
      if (parent instanceof Applet)
	((Applet)parent).showStatus ("Error copying to Virtual Clipboard: " + e.getMessage ());
    }
}

public final void loadFromFile ()
{
  Frame ff = new Frame ();
  FileDialog fd;
  try
    {
      try
	{
	  PrivilegeManager.enablePrivilege("UniversalFileAccess");
	}
      catch (Throwable t)
	{
	  throw new Exception (t.toString());
	}
      fd = new FileDialog (ff, "Load window from a file", FileDialog.LOAD);
      fd.show ();
      if (!(fd.getFile() == null || fd.getFile().equals ("")))
	{
	  String fname = fd.getDirectory () + fd.getFile ();
	  String s;
	  FileInputStream f;
	  StringTokenizer tok;
	  try
	    {
	      f = new FileInputStream (fname);
	    }
	  catch (Exception e)
	    {
	      throw new Exception ("Unable to open file '" + fname + "':\n" + e.toString());
	    }
				// read the opened file stream
	  
	  try
	    {
	      int count = f.available ();
	      byte bLoadBuf[] = new byte[count];
	      f.read (bLoadBuf);
	      f.close ();
	      s = new String (bLoadBuf, 0);
	    }
	  catch (Exception e1)
	    {
	      try
		{
		  f.close ();
		}
	      catch (Exception e2)
		{
		}
	      throw new Exception ("Error while reading file '" + fname + "':\n" + e1.toString());
	    }
	  setText (s);
	}
    }
  catch (Exception e)
    {
      if (parent instanceof Applet)
	((Applet)parent).showStatus (e.toString());
    }
}

public final void saveToFile ()
{
  Frame ff = new Frame ();
  FileOutputStream fosSave;
  try
    {
      try
	{
	  PrivilegeManager.enablePrivilege("UniversalFileAccess");
	}
      catch (Throwable t)
	{
	  throw new Exception (t.toString());
	}
      FileDialog fd = new FileDialog (ff, "Save window to file", FileDialog.SAVE);
      fd.show ();
      String fname = fd.getFile();
      if (fname == null || fname.equals (""))
	return;
      fname = fd.getDirectory () + fname;
      
      try
	{
	  fosSave = new FileOutputStream(fname);
	}
      catch (Exception e)
	{
	  throw new Exception ("Unable to create file '" + fname +"':\n  " + e.toString() );
	}
      try 
	{	  
	  String s = getText ();
	  int count =  s.length ();
	  byte bSaveBuf[] = new byte[count];
	  s.getBytes (0, count, bSaveBuf, 0);
	  fosSave.write (bSaveBuf);
	}
      catch (Exception e)
	{
	  try
	    {
	      fosSave.close ();
	    }
	  catch (Exception e2)
	    {
	    }
	  throw new Exception ("Unable to write to file '" + fname + "':\n  " + e.toString());
	}
      try
	{
	  fosSave.close ();
	}
      catch (Exception e)
	{
	  throw new Exception ("Unable to close file '" + fname + "':\n  " + e.toString());
	}
    }
  catch (Exception e)
    {
      if (parent instanceof Applet)
	((Applet)parent).showStatus (e.toString());
    }
}

public void setTitle (String t)
{
  lab.setText (" " + t + " ");
}

}
