package mylib;

import java.applet.*;

public final class ProgressBar
{
  Applet host;
  String message;
  int totalCount;
  int countSoFar;
  long startTime;			// when did this task start
  long prevTime;			// when was progress last shown
  int lastShown;			// what progress proportion was last displayed

  public ProgressBar (Applet parent, String s)
    {
      host = parent;
      message = s;
      countSoFar = 0;
      totalCount = 100;
      prevTime = startTime = System.currentTimeMillis ();
	
      displayProgress ();
    }

  public void SetTotalCount ( int t)
    {
      totalCount = t;
      countSoFar = 0;
      startTime = System.currentTimeMillis ();
    }

  public void SetMessage (String s)
    {
      message = s;
    }

  public void ShowProgress ()
    {

      ++countSoFar;
      displayProgress ();
    }


  public void ShowProgress (int c)
    {

      countSoFar += c;
      displayProgress ();
    }

  public void ShowProgress (double p)
    {

      countSoFar = (int) (p * totalCount);
      displayProgress ();
    }

  private void displayProgress ()
    {
      int progNow = countSoFar * 100 / totalCount;
      long timeNow = System.currentTimeMillis ();
      String msg;

      // show progress if the % count has gone up by 5 or if at least 5 seconds
      // have gone by since the last update
		
      if (progNow - lastShown > 5 || timeNow - prevTime > 5000)
	{
	  msg = "Running " + message + ": " + 
	    String.valueOf (100 * countSoFar / totalCount) + "% finished.";

	  if (countSoFar > 0)
	    {
	      long soFar = timeNow - startTime;
	      long left = (totalCount - countSoFar) * (soFar) / countSoFar;

	      msg = msg + "   Time Elapsed: " + String.valueOf (soFar / 60000) + "m" + String.valueOf ((soFar / 1000) % 60) + "s";
	      msg = msg + "  Left: " + String.valueOf (left / 60000) + "m" + String.valueOf ((left / 1000) % 60) + "s";
	    }

	  host.showStatus (msg);
	  lastShown = progNow;
	  prevTime = timeNow;
	}
    }

  public void finalize ()
    {
    }
}
