package mylib;
import java.util.StringTokenizer;
/*
 *
 * SparseVector
 *
 */
public class SparseVector 
{
	// a vector of (index, value) pairs

	boolean isSorted;		// true if the index array is in ascending order
	public int index[];
	public float value[];

	public SparseVector ()
	{
		index = null;
		value = null;
		isSorted = true;
	}

	public SparseVector (int n)
	{
		index = new int[n];
		value = new float[n];
	}


	public SparseVector (int index[], float value[])
	{
		this.index = index;
		this.value = value;
		checkIfSorted ();
	}


	public SparseVector (String s) throws Exception
	{
		// create a SparseVector from a string consisting of 
		// a sequence of numbers with indices in even positions (0, 2, 4, ...)
		// and values in odd positions (1, 3, 5, ...)

		StringTokenizer tok = new StringTokenizer (s, "\t\n\r ");
		int n = tok.countTokens () / 2;
		index = new int[n];
		value = new float[n];
		boolean gettingIndex = true;
		try 
		{
			for (int i=0; i < n; ++i)
			{
				gettingIndex = true;
				index[i] = Integer.parseInt (tok.nextToken ());
				gettingIndex = false;
				value[i] = new Float (tok.nextToken ()).floatValue ();
			}
			checkIfSorted ();
		}
		catch (Exception e)
		{
			throw new Exception ("While reading character vector, expected a" + (gettingIndex ? "n integer" : " number") + " but got: " + e.getMessage ());
		}
	}


	void checkIfSorted ()
	{
		if (index == null)
			return;
		isSorted = true;
		for (int i=1; i < index.length; ++i)
		{
			if (index[i] < index[i-1])
			{
				isSorted = false;
				return;
			}
		}
	}

	public void sort ()
	{
		// bubble sort the array, if necessary

		if (isSorted || index == null || value == null)
			return;

		for (int i=0; i < index.length-1; ++i )
		{
			for (int j=i; j < index.length;  ++j)
			{
				if (index[i] > index[j])
				{
					int t=index[i];
					index[i] = index[j];
					index[j] = t;
					float t2 = value[i];
					value[i] = value[j];
					value[j] = t2;
				}
			}
		}
		isSorted = true;
	}

	public float getValue (int i)
	{
		if (index == null || value == null)
			return 0;

		if (isSorted)
		{
			// binary search for the correct index
			int j1 = 0;
			int j2 = index.length - 1;
			for (;;)
			{
				if (i < index[j1] || i > index[j2])
					return 0;
				if (i == index[j1])
					return value[j1];
				if (i == index[j2])
					return value[j2];
				if (j1 == j2)
					return 0;
				int jm = (j1 + j2 + 1) / 2;
				if (i >= index[jm])
					j1 = jm;
				else
					j2 = jm;
			}
		}
		else
		{
			// search the index list
			for (int j=0; j < index.length; ++j)
				if (index[j] == i)
					return value[j];
			return 0;
		}
	}
}








